@@ -2,9 +2,7 @@ from pathlib import Path
 import logging
 
 from .opr3 import Data
-# from .helper import attributes, objects, explain
-# from .basequery.dissociated import *
-# from .readquery import *
+from .readquery import *
 
 from .__version__ import __version__
 
