@@ -36,6 +36,7 @@ import numpy as np
 from weaveio.config_tables import progtemp_config
 from weaveio.hierarchy import Hierarchy, Multiple, OneOf, Optional
 
+
 class ArrayHolder(Hierarchy):
     is_template = True
 
@@ -168,9 +169,9 @@ class Fibre(Hierarchy):
 
 class Subprogramme(Hierarchy):
     """
-    A submitted programme of observation which was written by multiple surveys.
+    A submitted programme of observation which was written by a survey.
     """
-    parents = [Multiple(Survey)]
+    parents = [Multiple(Survey, maxnumber=5)]
     factors = ['name']
     idname = 'id'
 
@@ -248,12 +249,12 @@ class OBSpec(Hierarchy):
     """
     When an xml observation specification is submitted to WEAVE, an OBSpec is created containing all
     the information about when and how to observe.
-    When actually observing them, an "OB" is create with its own unique obid.
+    When actually observing them, an "OB" is created with its own unique obid.
     """
     singular_name = 'obspec'
     factors = ['title']
-    parents = [Obstemp, Progtemp, Multiple(FibreTarget, one2one=True),  # each OB has a defined number of fibre_targets
-               Multiple(SurveyCatalogue), Multiple(Subprogramme), Multiple(Survey)]
+    parents = [Obstemp, Progtemp, Multiple(FibreTarget, maxnumber=1000, one2one=True),  # each OB has a defined number of fibre_targets
+               Multiple(SurveyCatalogue, maxnumber=10), Multiple(Subprogramme, maxnumber=10), Multiple(Survey, maxnumber=5)]
     idname = 'xml'  # this is CAT-NAME in the header not CATNAME, annoyingly no hyphens allowed
 
 
@@ -297,12 +298,14 @@ class Spectrum(Hierarchy):
 
 class Spectrum1D(Spectrum):
     is_template = True
-    parents = [WavelengthHolder]
+    children = [OneOf(WavelengthHolder, one2one=True)]
     products = ['flux', 'ivar']
+    plural_name = 'spectra1d'
 
 
 class Spectrum2D(Spectrum):
     is_template = True
+    plural_name = 'spectra2d'
 
 
 class Run(Hierarchy):
