@@ -1,13 +1,18 @@
 import logging
 import time
+from collections import defaultdict
 from contextlib import contextmanager
 from typing import Tuple, List, Dict, Any, Union, TYPE_CHECKING
+from warnings import warn
 
+import networkx as nx
 from networkx import NetworkXNoPath, NodeNotFound
+from weaveio.hierarchy import Hierarchy
 
 from .exceptions import AmbiguousPathError, CardinalityError, DisjointPathError, AttributeNameError
 from .parser import QueryGraph
-from .results import Table
+from .results import Table, AstropyTable
+from ..path_finding import collapse_classes_to_superclasses
 
 if TYPE_CHECKING:
     from .objects import ObjectQuery, Query, AttributeQuery
@@ -25,21 +30,43 @@ class BaseQuery:
     one_row = False
     one_column = False
 
+    def _debug_output(self):
+        n = self._precompile()
+        c = n._to_cypher()
+        return '\n'.join(c[0]), c[1]
+
     def raise_error_with_suggestions(self, obj, exception: Exception):
         self._data.autosuggest(obj, self._obj, exception)
         raise exception
 
     def __repr__(self):
-        return f'<{self.__class__.__name__}({self._previous._obj}-{self._obj})>'
+        try:
+            return f'<{self.__class__.__name__}({self._previous._obj}-{self._obj})>'
+        except AttributeError:
+            return f'<{self.__class__.__name__}({self._obj})>'
 
     def _precompile(self) -> 'BaseQuery':
         return self
 
+    def _prepare_parameters(self, lines):
+        params = {}
+        # aliases = self._G.get_unwind_variables(self._node)
+        for k, v in self._G.parameters.items():
+            if any(k in l for l in lines):
+                if isinstance(v, AstropyTable):
+                    cols = [c for c in v.colnames if any(c in l for l in lines)]
+                    assert len(cols), "No columns are used, but table is used. This should not happen"
+                    params[k] = v[cols]
+                else:
+                    params[k] = v
+        return params
+
     def _to_cypher(self) -> Tuple[List[str], Dict[str, Any]]:
         """
         returns the cypher lines, cypher parameters, names of columns, expect_one_row, expect_one_column
         """
-        return self._G.cypher_lines(self._node), self._G.parameters
+        lines = self._G.cypher_lines(self._node)
+        return lines, self._prepare_parameters(lines)
 
     def _compile(self) -> Tuple['BaseQuery', Tuple[List[str], Dict[str, Any]]]:
         with logtime('compiling'):
@@ -49,19 +76,25 @@ class BaseQuery:
     def _execute(self, skip=0, limit=None):
         with logtime('executing'):
             new, (lines, params) = self._compile()
+            cypher = '\n'.join(lines)
             if skip > 0:
-                lines.append(f"SKIP {skip}")
+                cypher += f"\nSKIP {skip}"
             if limit is not None:
-                lines.append(f"LIMIT {limit}")
-            cursor = new._data.graph.execute('\n'.join(lines), **{k.replace('$', ''): v for k,v in params.items()})
+                cypher += f"\nLIMIT {limit}"
+            cursor = new._data.graph.execute(cypher, **{k.replace('$', ''): v for k,v in params.items()})
             return cursor, new
 
     def _iterate(self, skip=0, limit=None):
         cursor, new = self._execute(skip, limit)
         rows = new._data.rowparser.iterate_cursor(cursor, new._names, new._is_products, True)
         with logtime('total streaming (by iteration)'):
-            for row in rows:
+            for i, row in enumerate(rows):
                 yield new._post_process_row(row)
+        if limit == i:
+            warn(f"This query has been capped to {limit} rows but the result is larger than that. "
+                 f"Consider using the `limit` parameter to better limit the result or set `limit` to None to get "
+                 f"the full result (although this is not recommended).")
+
 
     def __iter__(self):
         yield from self._iterate()
@@ -83,11 +116,18 @@ class BaseQuery:
             return row[row.colnames[0]]
         return row
 
-    def __call__(self, skip=0, limit=None, **kwargs):
+    def __call__(self, skip=0, limit=1000, **kwargs):
         tbl, new = self._to_table(skip, limit)
         tbl = new._post_process_table(tbl)
         if limit == 1:
             return tbl[0]
+        try:
+            if limit == len(tbl):
+                warn(f"This query has been capped to {limit} rows but the result is larger than that. "
+                     f"Consider using the `limit` parameter to better limit the result or set `limit` to None to get "
+                     f"the full result (although this is not recommended).")
+        except TypeError:
+            pass
         return tbl
 
 
@@ -122,7 +162,8 @@ class BaseQuery:
         else:
             self._start = start
         if self._obj is not None:
-            self._obj = self._normalise_object(self._obj)[0]
+            if not self._obj.startswith('_'):  # system name for other purposes
+                self._obj = self._normalise_object(self._obj)[0]
         self._names = [] if names is None else names
         self._is_products = [False]*len(self._names) if is_products is None else is_products
         self.attrs = attrs
@@ -131,43 +172,45 @@ class BaseQuery:
     def _get_object_of(self, maybe_attribute: str) -> Tuple[str, bool, bool]:
         """
         Given a str that might refer to an attribute, return the object that contains that attribute
-        and also whether the attribute and the object are singular
-        inferred obj | attr
-            s        |  s      (i.e. run.obid -> run.ob.obid)
-            p        |  p      (i.e. ...-> run.l1specs.snrs) - cannot be inferred just from a name
-            s        |  p      (i.e. run.obids -> run.ob.obids)
-            p        |  s      (i.e. run.snr -> run.l1spec.snr) - is not allowed
-         :return: obj, obj_is_singular, attr_is_singular
+        and also whether the attribute and the object are singular.
+        plurality table:
+            inferred obj | attr
+                s        |  s      (i.e. run.obid -> run.ob.obid)
+                p        |  p      (i.e. ...-> run.l1specs.snrs) - cannot be inferred just from a name
+                s        |  p      (i.e. run.obids -> run.ob.obids)
+                p        |  s      (i.e. run.snr -> run.l1spec.snr) - is not allowed
+        attribute storage scenarios:
+            * attribute is contained in this object - return self
+            * attribute is contained in exactly one object - return object
+            * attribute is contained in more than one object and ...
+                * any of those objects don't have the same inherited class - raise AmbiguousError
+                * only subclass objects and all subclass objects from a superclass contain the attribute - return the superclass
+                * only subclass objects and not all subclass objects from a superclass contain the attribute - return the superclass
+        If a singular path exists, use that
+        :return: obj, obj_is_singular, attr_is_singular
         """
         single_name = self._data.singular_name(maybe_attribute)
-        hs = {h.__name__ for h in self._data.factor_hierarchies[single_name]}
-        if self._obj in hs:
-            return self._obj, True, self._data.is_singular_name(maybe_attribute)
-        # find the objects that are linked to this one and sort them by the amount of data they require
-        new_hs = []
-        for H in hs:
-            for h in self._data.expand_template_object(H):
-                try:
-                    level = self._get_path_to_object(h, False)[-1]
-                    new_hs.append((h, level))
-                except (NetworkXNoPath, NodeNotFound):
-                    pass
-        new_hs.sort(key=lambda x: x[1])
-        lowest_level = min(list(zip(*new_hs))[1])
-        hs = [h for h, l in new_hs if l == lowest_level]  # pick the objs with the smallest data requirement
-        if len(hs) > 1:
-            names = [self._data.singular_name(h) for h in hs]
-            raise AmbiguousPathError(f"There are multiple attributes called {maybe_attribute} with the following parent objects: {names}."
-                                     f" Please be specific e.g. `{names[0]}.{maybe_attribute}`")
-        try:
-            obj = hs[0]
-        except IndexError:
-            raise AttributeNameError(f"{self._obj} has no access to an attribute called {maybe_attribute}")
+        if self._obj is not None:
+            if self._data.class_hierarchies[self._obj] in self._data.factor_hierarchies[single_name]:
+                return self._obj, True, self._data.is_singular_name(maybe_attribute)
+        want_single = single_name == maybe_attribute
+        # try to get singular paths
+        hs = {h for h in self._data.factor_hierarchies[single_name] if self._has_path_to_object(h, True)}
+        if len(hs) == 0: # if no singular paths then get all paths
+            hs = {h for h in self._data.factor_hierarchies[single_name] if self._has_path_to_object(h, want_single)}
+        hs = set(collapse_classes_to_superclasses(self._data.hierarchy_graph, hs))
+        if len(hs) == 1:
+            obj = hs.pop().__name__
+        elif not hs:
+            raise AmbiguousPathError(f"There is no path to any object that contains `{single_name}`. {self._data.factor_hierarchies[single_name]} are not reachable")
+        else:
+            raise AmbiguousPathError(f"`{single_name}` is contained within more than one unrelated object {[h.__name__ for h in hs]}. Be specific.")
         if self._obj is None:
             return obj, True, False
         if not self._data.is_factor_name(maybe_attribute):
             raise ValueError(f"{maybe_attribute} is not a valid attribute name")
-        path, obj_is_singular, _ = self._data.path_to_hierarchy(self._obj, obj, False)
+        _, obj_is_singulars = self._get_path_to_object(obj, False)
+        obj_is_singular = len(obj_is_singulars) == 1 and obj_is_singulars[0]
         attr_is_singular = self._data.is_singular_name(maybe_attribute)
         if obj_is_singular and attr_is_singular:
             pass  # run.obid -> run.ob.obid
@@ -218,8 +261,16 @@ class BaseQuery:
     def _spawn(cls, parent: 'BaseQuery', node, obj=None, index_node=None, single=False, *args, **kwargs):
         return cls(parent._data, parent._G, node, parent, obj, parent._start, index_node, single, *args, **kwargs)
 
-    def _get_path_to_object(self, obj, want_single) -> Tuple[str, bool, int]:
-        return self._data.path_to_hierarchy(self._obj, obj, want_single)
+    def _get_path_to_object(self, obj, want_single) -> Tuple[List[str], List[bool]]:
+        return self._data.paths_to_hierarchy(self._obj, obj, want_single)
+
+    def _has_path_to_object(self, obj, want_single) -> bool:
+        if self._obj is None:
+            return True
+        try:
+            return len(self._get_path_to_object(obj, want_single)[0]) > 0
+        except NetworkXNoPath:
+            return False
 
     def _slice(self, slc):
         """
