@@ -9,22 +9,26 @@ with waiting,
     much better!
 it treats two chained expressions as one action
 """
-from typing import List, TYPE_CHECKING, Union, Tuple, Dict, Any
+import collections
+from copy import copy
+from typing import List, TYPE_CHECKING, Union, Tuple, Dict, Any, Iterable
 
 from networkx import NetworkXNoPath
 
 from .base import BaseQuery
 from .exceptions import CardinalityError, AttributeNameError, UserError
 from .parser import QueryGraph, ParserError
-from .utilities import is_regex, dtype_conversion
-from ..opr3.hierarchy import Exposure
+from .utilities import is_regex, dtype_conversion, mask_infs
+from astropy.table import Table
+
 
 if TYPE_CHECKING:
     from weaveio.data import Data
 
 
 class GenericObjectQuery(BaseQuery):
-    pass
+    def _ipython_key_completions_(self):
+        return [i.strip('"') for i in self.__dir__()]
 
 
 def process_names(given_names, attrs: List['AttributeQuery']) -> List[str]:
@@ -42,7 +46,19 @@ def process_names(given_names, attrs: List['AttributeQuery']) -> List[str]:
             names[i] = f"{n}{j}"
     return [a._factor_name for a in attrs]  # todo: this could be better
 
+
 class ObjectQuery(GenericObjectQuery):
+    def __dir__(self) -> List[str]:
+        h = self._data.class_hierarchies[self._obj]
+        singulars = set(self._data.hierarchy_graph.singular.neighbors(h))
+        all_ = set(self._data.hierarchy_graph.neighbors(h))
+        neighbors = [i.singular_name for i in singulars]
+        neighbors += [i.plural_name for i in all_ - singulars]
+        neighbors += list(h.relative_names.keys())
+        neighbors += h.products_and_factors
+        neighbors = [i if '[' not in i else f'"{i}"' for i in neighbors]
+        return [i.lower() for i in neighbors]
+
     def _precompile(self) -> 'TableQuery':
         """
         If the object contains only one factor/product and defines no parents/children, return that
@@ -98,14 +114,15 @@ class ObjectQuery(GenericObjectQuery):
         plural = self._data.plural_name(obj)
         err_msg = f"There is no singular {obj} relative to {self._obj} try using its plural {plural}"
         try:
-            path, single, _ = self._get_path_to_object(obj, want_single)
+            paths, singles = self._get_path_to_object(obj, want_single)
+            single = len(singles) == 1 and singles[0]
         except NetworkXNoPath:
             if not want_single:
                 err_msg = f"There is no {obj} relative to {self._obj}"
             raise CardinalityError(err_msg)
         if not single and want_single:
             raise CardinalityError(err_msg)
-        n = self._G.add_traversal(self._node, path, obj, single)
+        n = self._G.add_traversal(self._node, paths, obj, single)
         return ObjectQuery._spawn(self, n, obj, single=want_single)
 
     def _traverse_by_object_index(self, obj, index):
@@ -117,7 +134,7 @@ class ObjectQuery(GenericObjectQuery):
         e.g. `obs[1234]` filters to the single ob with obid=1234
         """
         param = self._G.add_parameter(index)
-        path, single, _ = self._get_path_to_object(obj, False)
+        path, single = self._get_path_to_object(obj, False)
         travel = self._G.add_traversal(self._node, path, obj, single)
         i = self._G.add_getitem(travel, 'id')
         eq, _ = self._G.add_scalar_operation(i, f'{{0}} = {param}', f'id={index}')
@@ -125,12 +142,12 @@ class ObjectQuery(GenericObjectQuery):
         return ObjectQuery._spawn(self, n, obj, single=True)
 
     def _traverse_by_object_indexes(self, obj, indexes: List):
+        path, single = self._get_path_to_object(obj, False)
         param = self._G.add_parameter(indexes)
-        path, single, _ = self._get_path_to_object(obj, False)
         one_id = self._G.add_unwind_parameter(self._node, param)
         travel = self._G.add_traversal(self._node, path, obj, single, one_id)
         i = self._G.add_getitem(travel, 'id')
-        eq, _ = self._G.add_combining_operation('{0} = {1}', 'ids', i, one_id)
+        eq, _ = self._G.add_combining_operation('{0} = {1}', 'ids', i, one_id, wrt=travel)
         n = self._G.add_filter(travel, eq, direct=True)
         return ObjectQuery._spawn(self, n, obj, single=True)
 
@@ -162,7 +179,7 @@ class ObjectQuery(GenericObjectQuery):
         r._index_node = self._node
         return r
 
-    def _make_table(self, *items):
+    def _make_table(self, items):
         """
         obj['factor_string', AttributeQuery]
         obj['factora', 'factorb']
@@ -171,6 +188,10 @@ class ObjectQuery(GenericObjectQuery):
         """
         attrs = []
         for item in items:
+            if isinstance(item, dict):
+                if len(item) > 2:
+                    raise ValueError(f"Can only have one key-value pair per item, got {item}")
+                name, item = copy(item).popitem()
             if isinstance(item, ObjectQuery):
                 item = item._precompile()
             if isinstance(item, AttributeQuery):
@@ -185,9 +206,9 @@ class ObjectQuery(GenericObjectQuery):
                 attrs.append(new)
         force_plurals = [not a._single for a in attrs]
         is_products = [a._is_products[0] for a in attrs]
-        n = self._G.add_results_table(self._node, [a._node for a in attrs], force_plurals, dropna=[self._node])
-        names = process_names([i if isinstance(i, str) else None for i in items], attrs)
-        return TableQuery._spawn(self, n, names=names, is_products=is_products, attrs=attrs)
+        n = self._G.add_results_table(self._node, [a._node for a in attrs], force_plurals)
+        names = process_names([i if isinstance(i, str) else list(i.keys())[0] if isinstance(i, dict) else None for i in items], attrs)
+        return TableQuery._spawn(self, n, names=names, is_products=is_products, attr_queries=attrs)
 
     def _traverse_to_relative_object(self, obj, index):
         """
@@ -202,7 +223,7 @@ class ObjectQuery(GenericObjectQuery):
         """
         obj, obj_singular = self._normalise_object(obj)
         singular_name = self._data.singular_name(index)
-        path, single, _ = self._get_path_to_object(obj, False)
+        path, single = self._get_path_to_object(obj, False)
         if not single and singular_name == index:
             raise SyntaxError(f"Relative index `{index}` is plural relative to `{self._obj}`.")
         n = self._G.add_traversal(self._node, path, obj, False)
@@ -223,11 +244,21 @@ class ObjectQuery(GenericObjectQuery):
         raise NotImplementedError
 
     def _getitems(self, items, by_getitem):
-        if not all(isinstance(i, (str, float, int, AttributeQuery)) for i in items):
+        # can be called with items signifying columns or where items are indexes, need to distinguish here
+        # standardise item list
+        if isinstance(items, dict):
+            items = [{k: v} for k, v in items.items()]
+        else:
+            items = list(iter(items))
+        items = [[{k: v} for k, v in item.items()] if isinstance(item, dict) else [item] for item in items]
+        items = [i for item in items for i in item]
+        values = [list(i.values())[0] if isinstance(i, dict) else i for i in items]
+        # names = [list(i.keys())[0] if isinstance(i, dict) else i for i in items]
+        if not all(isinstance(i, (str, float, int, AttributeQuery)) for i in values):
             raise TypeError(f"Cannot index by non str/float/int/AttributeQuery values")
-        if all(self._data.is_valid_name(i) or isinstance(i, AttributeQuery) for i in items):
-            return self._make_table(*items)
-        if any(self._data.is_valid_name(i) for i in items):
+        if all(self._data.is_valid_name(i) or isinstance(i, AttributeQuery) for i in values):
+            return self._make_table(items)
+        if any(self._data.is_valid_name(i) for i in values):
             raise SyntaxError(f"You may not mix filtering by id and building a table with attributes")
         # go back and be better
         return self._previous._traverse_by_object_indexes(self._obj, items)
@@ -236,7 +267,7 @@ class ObjectQuery(GenericObjectQuery):
         """
         item can be an id, a factor name, a list of those, a slice, or a boolean_mask
         """
-        if isinstance(item, (tuple, list)):
+        if isinstance(item, collections.Iterable) and not isinstance(item, (str, BaseQuery)):
             return self._getitems(item, by_getitem)
         elif isinstance(item, slice):
             return self._slice(item)
@@ -254,7 +285,7 @@ class ObjectQuery(GenericObjectQuery):
             try:
                 return self._select_or_traverse_to_attribute(item)
             except (KeyError, ValueError):
-                if '.' in item:  # split the parts and parse
+                if '.' in item and not (item.endswith('.fit') or item.endswith('.fits')):  # split the parts and parse
                     try:
                         obj, attr = item.split('.')
                         return self.__getitem__(obj).__getitem__(attr)
@@ -288,9 +319,15 @@ class ObjectQuery(GenericObjectQuery):
             self._data.autosuggest(item, self._obj, e)
 
     def __getattr__(self, item):
+        if isinstance(item, str):
+            if item.startswith('_'):
+                raise super(ObjectQuery, self).__getattribute__(item)
         return self._getitem_handled(item, False)
 
     def __getitem__(self, item):
+        if isinstance(item, str):
+            if item.startswith('_'):
+                raise KeyError
         return self._getitem_handled(item, True)
 
     def __eq__(self, other):
@@ -307,6 +344,59 @@ class ObjectQuery(GenericObjectQuery):
         return self._select_attribute(i, True).__eq__(other)
 
 
+class TableVariableQuery(ObjectQuery):
+    def __dir__(self) -> List[str]:
+        return self._table.colnames
+
+    def __init__(self, data: 'Data', G: QueryGraph = None, node=None, previous: Union['Query', 'AttributeQuery', 'ObjectQuery'] = None,
+                 obj: str = None, start: 'Query' = None, index_node=None, single=False, names=None, is_products=None, attrs=None,
+                 dtype=None, table=None, *args, **kwargs) -> None:
+        super().__init__(data, G, node, previous, obj, start, index_node, single, names, is_products, attrs, dtype, *args, **kwargs)
+        self._table = table
+
+    def _precompile(self) -> 'TableQuery':
+        return self._get_all_factors_table()
+
+    def _get_all_factors_table(self) -> 'TableQuery':
+        return self._make_table(self._table.colnames)
+
+    def _select_all_attrs(self) -> 'TableQuery':
+        return self._get_all_factors_table()
+
+    def _traverse_to_generic_object(self):
+        raise NotImplementedError
+
+    def _traverse_to_specific_object(self, obj, want_single):
+        raise NotImplementedError
+
+    def _traverse_by_object_index(self, obj, index):
+        raise NotImplementedError
+
+    def _traverse_by_object_indexes(self, obj, indexes: List):
+        raise NotImplementedError
+
+    def _select_product(self, attr, want_single):
+        raise NotImplementedError
+
+    def _select_attribute(self, column_name, want_single):
+        if column_name not in self._table.colnames:
+            raise KeyError(f"`{column_name}` is not a column in the table `{self}`")
+        n = self._G.add_getitem(self._node, column_name)
+        return AttributeQuery._spawn(self, n, single=want_single, factor_name=column_name)
+
+    def _select_or_traverse_to_attribute(self, attr):
+        return self._select_attribute(attr, True)
+
+    def _traverse_to_relative_object(self, obj, index):
+        raise NotImplementedError
+
+    def _filter_by_relative_index(self):
+        raise NotImplementedError
+
+    def __eq__(self, other):
+        raise NotImplementedError
+
+
 class Query(GenericObjectQuery):
     def __init__(self, data: 'Data', G: QueryGraph = None, node=None, previous: 'BaseQuery' = None, obj: str = None, start=None) -> None:
         super().__init__(data, G, node, previous, obj, start, 'start')
@@ -324,7 +414,10 @@ class Query(GenericObjectQuery):
     def _traverse_by_object_index(self, obj, index):
         obj, single = self._normalise_object(obj)
         name = self._G.add_parameter(index)
-        travel = self._G.add_start_node(obj)
+        if self._node == 0:
+            travel = self._G.add_start_node(obj)
+        else:
+            travel = self._G.add_traversal(self._node, [], obj, single)
         i = self._G.add_getitem(travel, 'id')
         eq, _ = self._G.add_scalar_operation(i, f'{{0}} = {name}', f'id={index}')
         n = self._G.add_filter(travel, eq, direct=True)
@@ -333,8 +426,11 @@ class Query(GenericObjectQuery):
     def _traverse_by_object_indexes(self, obj, indexes: List):
         param = self._G.add_parameter(indexes)
         one_id = self._G.add_unwind_parameter(self._node, param)
-        travel = self._G.add_start_node(obj, one_id)
-        i = self._G.add_getitem(travel, 'id')
+        if self._node == 0:
+            travel = self._G.add_start_node(obj, one_id)
+        else:
+            travel = self._G.add_traversal(self._node, [], obj, False, one_id)
+        i = self._G.add_getitem(travel, self._data.class_hierarchies[obj].idname)
         eq, _ = self._G.add_combining_operation('{0} = {1}', 'ids', i, one_id)
         n = self._G.add_filter(travel, eq, direct=True)
         return ObjectQuery._spawn(self, n, obj, single=True)
@@ -374,7 +470,7 @@ class AttributeQuery(BaseQuery):
             n, wrt = self._G.add_combining_operation(op_string, 'op-filter', self._node, item._node)
         except ParserError:
             raise SyntaxError(f"You may not perform an operation on {self} and {item} since one is not an ancestor of the other")
-        return AttributeQuery._spawn(self, n, index_node=wrt, single=True, dtype=self.dtype)
+        return AttributeQuery._spawn(self, n, index_node=n, single=True, dtype=self.dtype)
 
     def _perform_arithmetic(self, op_string, op_name, other=None, expected_dtype=None, returns_dtype=None):
         """
@@ -396,11 +492,13 @@ class AttributeQuery(BaseQuery):
             # op_string = op_string.replace('{0}', f'to{expected_dtype}({{0}})')
             # op_string = op_string.replace('{1}', f'to{expected_dtype}({{1}})')  # both arguments
         if isinstance(other, BaseQuery):
+            op_string = op_string.format(mask_infs('{0}'), mask_infs('{1}'))
             try:
                 n, wrt = self._G.add_combining_operation(op_string, op_name, self._node, other._node)
             except ParserError:
                 raise SyntaxError(f"You may not perform an operation on {self} and {other} since one is not an ancestor of the other")
         else:
+            op_string = op_string.format(mask_infs('{0}'))
             n, wrt = self._G.add_scalar_operation(self._node, op_string, op_name)
         return AttributeQuery._spawn(self, n, index_node=wrt, single=True, dtype=returns_dtype)
 
@@ -506,7 +604,11 @@ class AttributeQuery(BaseQuery):
             index = self._G.start
         else:
             index = self._index_node
-        r = self._G.add_results_table(index, [self._node], [not self._single], dropna=[self._node])
+        if index == self._node:  # this happens if we've filtered an attribute by something
+            dropna = index
+        else:
+            dropna = None  # figure it out automatically
+        r = self._G.add_results_table(index, [self._node], [not self._single], dropna=dropna)
         a = AttributeQuery._spawn(self, r, self._obj, index, self._single, is_products=self._is_products,
                                   factor_name=self._factor_name)
         if index == 0:
@@ -532,10 +634,14 @@ class TableQuery(BaseQuery):
                  single=False, attr_queries=None, names=None, *args, **kwargs) -> None:
         super().__init__(data, G, node, previous, obj, start, index_node, single, names, *args, **kwargs)
         self._attr_queries = attr_queries
+        self._lookup = {k: v for k, v in zip(self._names, self._attr_queries)}
 
     def _aggregate(self, wrt, string_op, predicate=False, expected_dtype=None, returns_dtype=None, remove_infs=None):
         return self._previous._aggregate(wrt, string_op, predicate, expected_dtype, returns_dtype, remove_infs)
 
+    def __getitem__(self, item):
+        return self._lookup[item]
+
 
 class ListAttributeQuery(AttributeQuery):
     pass
