@@ -1,12 +1,15 @@
 from collections import defaultdict
 from typing import List, Tuple, Dict
+from pathlib import Path
+import warnings
 
 import networkx as nx
-from pathlib import Path
+import pandas as pd
+from astropy.table import Table
 
 from .utilities import mask_infs, remove_successive_duplicate_lines, dtype_conversion
-from .digraph import HashedDiGraph, plot_graph, add_start, add_traversal, add_filter, add_aggregation, add_operation, add_return, add_unwind, subgraph_view, get_above_state_traversal_graph, node_dependencies
-from .statements import StartingMatch, Traversal, NullStatement, Operation, GetItem, AssignToVariable, DirectFilter, CopyAndFilter, Aggregate, Return, Unwind, GetProduct, UnionTraversal
+from .digraph import HashedDiGraph, plot_graph, add_start, add_traversal, add_filter, add_aggregation, add_operation, add_return, add_unwind, subgraph_view, get_above_state_traversal_graph, node_dependencies, add_node_reference
+from .statements import StartingMatch, Traversal, NullStatement, Operation, GetItem, AssignToVariable, DirectFilter, CopyAndFilter, Aggregate, Return, Unwind, GetProduct, UnionTraversal, ApplyToList
 
 
 class ParserError(Exception):
@@ -253,6 +256,8 @@ class QueryGraph:
         return {d['statement']: (a, b) for a, b, d in self.G.edges(data=True) if 'statement' in d}
 
     def latest_shared_ancestor(self, *nodes):
+        if all(n == nodes[0] for n in nodes):
+            return nodes[0]
         return sorted(set.intersection(*[self.above_state(n, no_wrt=True) for n in nodes]), key=lambda n: len(nx.ancestors(self.traversal_G, n)))[-1]
 
     def latest_object_node(self, a, b):
@@ -344,14 +349,8 @@ class QueryGraph:
         statement = StartingMatch(node_type, unwound, self)
         return add_traversal(self.G, parent_node, statement, unwound=unwound)
 
-    def add_traversal(self, parent_node, path: str, end_node_type: str, single=False, unwound=None):
-        statement = Traversal(self.G.nodes[parent_node]['variables'][0], end_node_type, path, unwound, self)
-        return add_traversal(self.G, parent_node, statement, single=single, unwound=unwound)
-
-    def add_union_traversal(self, parent_node, paths: Dict[str, str], end_node_type: str, single=False, unwound=None):
-        if len(paths) == 1:
-            return self.add_traversal(parent_node, paths[list(paths.keys())[0]], end_node_type, single=single, unwound=unwound)
-        statement = UnionTraversal(self.G.nodes[parent_node]['variables'][0], paths, end_node_type, unwound, self)
+    def add_traversal(self, parent_node, paths: List[str], end_node_type: str, single=False, unwound=None):
+        statement = Traversal(self.G.nodes[parent_node]['variables'][0], end_node_type, paths, unwound, self)
         return add_traversal(self.G, parent_node, statement, single=single, unwound=unwound)
 
     def fold_to_cardinal(self, parent_node):
@@ -447,9 +446,14 @@ class QueryGraph:
         statement = FilterClass(self.G.nodes[parent_node]['variables'][0], predicate, self)
         return add_filter(self.G, parent_node, [predicate_node], statement)
 
-    def add_unwind_parameter(self, wrt_node, to_unwind):
-        statement = Unwind(wrt_node, to_unwind, 'unwound_'+to_unwind.replace('$', ''), self)
-        return add_unwind(self.G, wrt_node, statement)
+    def add_apply_to_list(self, parent_node, list_variable, apply_function, filter_function, *dependencies, put_null_in_empty=False):
+        dependencies = [self.fold_to_cardinal(d) for d in dependencies]
+        statement = ApplyToList(list_variable, [self.G.nodes[d]['variables'][0] for d in dependencies], apply_function, filter_function, self, put_null_in_empty)
+        return add_operation(self.G, parent_node, dependencies, statement), statement.output_variables[0]
+
+    def add_unwind_parameter(self, wrt_node, to_unwind, *dependencies):
+        statement = Unwind(wrt_node, to_unwind, 'unwound', self)
+        return add_unwind(self.G, wrt_node, statement, *dependencies)
 
     def collect_or_not(self, index_node, other_node, force_plural):
         """
@@ -477,7 +481,7 @@ class QueryGraph:
         else:
             return self.add_aggregation(other_node, shared, 'collect')
 
-    def add_results_table(self, index_node, column_nodes, force_plurals: List[bool], dropna: List):
+    def add_results_table(self, index_node, column_nodes, force_plurals: List[bool], dropna=None):
         # fold back column data into the index node
         column_nodes = [self.collect_or_not(index_node, d, p) for d, p in zip(column_nodes, force_plurals)] # fold back when combining
         deps = [self.G.nodes[d]['variables'][0] for d in column_nodes]
@@ -485,10 +489,13 @@ class QueryGraph:
             vs = self.G.nodes[index_node]['variables'][0]
         except IndexError:
             vs = None
-        try:
-            dropna = self.G.nodes[index_node]['variables'][0]
-        except IndexError:
-            dropna = None
+        if dropna is None:
+            try:
+                dropna = self.G.nodes[index_node]['variables'][0]
+            except IndexError:
+                dropna = None
+        else:
+            dropna = self.G.nodes[dropna]['variables'][0]
         statement = Return(deps, vs, dropna, self)
         return add_return(self.G, index_node, column_nodes, statement)
 
@@ -499,11 +506,25 @@ class QueryGraph:
         return add_return(self.G, self.start, column_nodes, statement)
 
     def add_parameter(self, value, name=None):
+        if isinstance(value, pd.DataFrame):
+            value = Table.from_pandas(value)
+        if isinstance(value, Table):
+            # limit column multidimensional sizes to 100
+            allowed_cols = [c for c in value.colnames if sum(value[c].shape) / len(value) <= 100]
+            if allowed_cols != value.colnames:
+                warnings.warn(f"Columns {set(value.colnames) - set(allowed_cols)} were dropped due to size limitations (>=100)")
+                value = value[allowed_cols]
+        if value in self.parameters.values():
+            return [k for k, v in self.parameters.items() if v == value][0]
         name = f'${name}' if name is not None else '$'
         varname = self.get_variable_name(name)
         self.parameters[varname] = value
         return varname
 
+    def add_previous_reference(self, parent_node, node_to_reference):
+        vars = self.G.nodes[node_to_reference]['variables']
+        return add_node_reference(self.G, NullStatement(vars, self), parent_node, node_to_reference)
+
     def restricted(self, result_node=None) -> HashedDiGraph:
         if result_node is None:
             return nx.subgraph_view(self.G)
@@ -521,25 +542,28 @@ class QueryGraph:
         return verify_traversal(graph, ordering)
 
     def cypher_lines(self, result):
-        import time
-        start_time = time.perf_counter()
-        ordering = self.traverse_query(result)
-        self.verify_traversal(result, ordering)
-        statements = []
-        for i, e in enumerate(zip(ordering[:-1], ordering[1:])):
-            try:
-                statement = self.G.edges[e]['statement'].make_cypher(ordering[:i+1])
-                if statement is not None:
-                    statements.append(statement)
-            except KeyError:
-                pass
-        end_time = time.perf_counter()
-        timed = end_time - start_time
-        return remove_successive_duplicate_lines(statements)
+        try:
+            cypher = self.G.nodes[result]['cypher']
+        except KeyError:
+            ordering = self.traverse_query(result)
+            self.verify_traversal(result, ordering)
+            statements = []
+            for i, e in enumerate(zip(ordering[:-1], ordering[1:])):
+                try:
+                    statement = self.G.edges[e]['statement'].make_cypher(ordering[:i+1])
+                    if statement is not None:
+                        statements.append(statement)
+                except KeyError:
+                    pass
+            cypher = remove_successive_duplicate_lines(statements)
+            self.G.nodes[result]['cypher'] = cypher
+        return cypher
 
     def node_is_null_statement(self, node):
         if self.node_holds_type(node, 'aggr'):
             return any(isinstance(d['statement'], NullStatement) for _, _, d in self.G.in_edges(node, data=True))
         return False
 
-
+    def get_unwind_variables(self, until_node):
+        variables = {d['statement'].parameter: d['statement'].output for *e, d in self.restricted(until_node).edges(data=True) if isinstance(d.get('statement', None), Unwind)}
+        return {p: v for k, v in variables.items() for p in sorted(self.parameters.keys(), key=len, reverse=True) if p in k}
